$ifndef _STRINGARRAY_INC
$define _STRINGARRAY_INC
' ------------------------------------------------------------------------------------------------------------
'| This source code is provided 'as is' and any express or implied warranties, including,                     |
'| but not limited to, the implied warranties of merchantability and fitness for a particular                 |
'| purpose are disclaimed.                                                                                    |
'|                                                                                                            |
'| In no event shall the author, Crownhill or contributers be liable for any direct, indirect, incidental,    |
'| special, exemplary, or consequential damages (including, but not limited to,  procurement of substitute    |
'| goods or services; loss of use, data, or profits;  or business interruption) however caused and on any     |
'| theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise)    |
'| arising in any way out of the use of this source code, even if advised of the possibility of such damage.  |
'|                                                                                                            |
'| Any views and conclusions contained in the source code and any documentation are those of the author's     |
'| and should not be interpreted as representing official policies,                                           |
'| either expressed or implied, of Crownhill Associates Ltd.                                                  |
' ------------------------------------------------------------------------------------------------------------
'
' Simulate a String array load and retrieve
' The macros use no variable RAM other that SFR registers
'
' For 18F devices only using the Proton compiler version 3.5.6.6 onwards
' Written by Les Johnson 20-01-14
'
'------------------------------------------------------------------------------------------
' Create a RAM string array
' Syntax:
'   Create_StringArray(Array Name , Amount of Strings, Size of each String)
'
' This will also create two constants:
' ArrayName#Amount holds the amount of strings
' ArrayName#Length holds the size of each string within the array
'
$define Create_StringArray(pName,pAmount,pLength)   '
Declare Warnings = Off                              '
    Dim pName[pAmount * (pLength + 1)] As Byte Heap '
    Symbol pName#Amount = pAmount                   '
    Symbol pName#Length = pLength + 1               '
Asm-                                                '
    pName#v(99) = pLength + 1                       '
Endasm-

'------------------------------------------------------------------------------------------
' Create a code memory string array
' Syntax:
'   Create_CodeStringArray(Array Name , Amount of Strings, Size of each String)
'
' This will also create a constant:
' ArrayName#Size holds the size of each string within the array
'
$define Create_CodeStringArray(pName,pAmount,pLength)            '
    Symbol pName#Length = pLength                                '
    $if pLength % 2 != 0                                         '
        $error "Code memory String length must be an even value" '
    $endif                                                       '
Asm-                                                             '
    pName#v(98) = pLength                                        '
Endasm-

'--------------------------------------------------------------------------------------

    GoTo _StringArray_Main_                             ' Jump over subroutines

'--------------------------------------------------------------------------------------
' Write to a RAM String array.
' The calculation for accessing the elements is: -
' FSR0 = pStringName
' PROD = pIndex * string's length
' FSR0 = FSR0 + PROD

' Syntax:-
' Write_StringArray pStringName, [pIndex], pCharacters

' Where: -
' pStringName is the name of the string array in question
' pIndex is the index value of the string array
' pCharacters are the characters to place into the array
'
Write_StringArray Macro- pStringName, pIndex, pCharacters

    #if(Prm_Count != 3)                                     ' Ensure the correct amount of parameters
        #error "Incorrect amount of parameters for Macro Write_StringArray. 3 Parameters required"
        Exitm
    #endif
    #if(Prm_1 != String) && (Prm_1 != Byte_Array_Ptr)       ' Ensure the first parameter is a string or a byte array
        #error "Parameter 1 must be a String Array for Macro Write_StringArray"
        Exitm
    #endif

    #if((Prm_2 == Num8) || (Prm_2 == Num16) || (Prm_2 == Num32))
        Lfsr 0,(pStringName + (pIndex * pStringName#v(99))) ' Point FSR0L/H to the address of the string array
    #endif

    #if((Prm_2 == Byte) || (Prm_2 == Word) || (Prm_2 == Dword) || (Prm_2 == SByte) || (Prm_2 == SWord) || (Prm_2 == SDword))
        Byte_Wreg pIndex
        Lfsr 0,(pStringName)                                ' Point FSR0L/H to the address of the string array
        Mullw (pStringName#v(99))                           ' \
        Movf PRODL,w                                        ' |
        Addwf FSR0L,f                                       ' | Calculate the position of the string array within the main String
        Movf PRODH,w                                        ' |
        Addwfc FSR0H,f                                      ' /
    #endif
'
' FSR0L and FSR0H are now pointing to the correct string within the array based on pIndex and the length of each string
'
    #if((Prm_3 == Char) || (Prm_3 == Label))                ' Is the parameter a quoted string of characters or a label?
        Movlw (pCharacters & 255)                           ' \
        Movwf TBLPTRL                                       ' | Point TBLPTRL/H to the code memory holding the string of characters
        Movlw ((pCharacters >> 8) & 255)                    ' |
        Movwf TBLPTRH                                       ' /
        Clrf EECON1                                         ' \ Setup For reading from code memory
        Bsf EECON1,EEPGD                                    ' /
        Movlw (pStringName#v(99))                           ' Create a loop for all the characters in the string array part
        Tblrd*+                                             ' Read then increment code memory pointer
        Movf TABLAT,f                                       ' \
        Bz $ + 10                                           ' / Exit if a null found
        Movff TABLAT,POSTINC0                               ' Place the character from code memory to RAM
        Decfsz WREG,f                                       ' Next character
        Bra $ - 12                                          ' Until all characters transfered
    #endif
    #if(Prm_3 == String)                                    ' Is the parameter another string?
        Lfsr 1,(pCharacters)                                ' Point FSR1L/H to the address of the source string
        Movlw (pStringName#v(99))                           ' Create a loop for all the characters in the string array part
        Movf INDF1,f                                        ' \
        Bz $ + 10                                           ' / Exit if a null found
        Movff POSTINC1,POSTINC0                             ' Place the character from source string to dest string
        Decfsz WREG,F                                       ' Next character
        Bra $ - 10                                          ' Until all characters transfered
    #endif
    #if((Prm_3 == Word) || (Prm_3 == Dword) || (Prm_3 == SWord) || (Prm_3 == SDword))   ' Is the parameter a word or dword?
        Movff (pCharacters,FSR0L)                           ' \
        Movff (pCharacters + 1,FSR0H)                       ' / Point FSR1L/H to the address of the source string
        Movlw (pStringName#v(99))                           ' Create a loop for all the characters in the string array part
        Movf INDF1,f                                        ' \
        Bz $ + 10                                           ' / Exit if a null found
        Movff POSTINC1,POSTINC0                             ' Place the character from source string to dest string
        Decfsz WREG,f                                       ' Next character
        Bra $ - 10                                          ' Until all characters transfered
    #endif
    #if(Prm_3 == Byte) || (Prm_3 == SByte)                  ' Is the parameter a byte
        Movff(pCharacters,INDF0)
    #endif
    #if((Prm_3 == Num8) || (Prm_3 == Num16) || (Prm_3 == Num32)) ' Is the parameter a constant?
        Num_SFR pCharacters,INDF0
    #endif

    Endm
'--------------------------------------------------------------------------------------
' Read from a RAM String array
' The calculation for accessing the elements is: -
' FSR0 = pStringName
' PROD = bIndex * string's length
' FSR0 = FSR0 + PROD

' Syntax:
' Dest_String = Read_StringArray, pStringLength, [pIndex]

' Where:-
' Dest_String is the destination string to hold the string extracted from the array
' pStringName is the name of the array in question
' pIndex is the index value of the string array
'
Read_StringArray Macro- pStringName, pIndex \Word

    #if(Prm_Count != 2)
        #error "Incorrect amount of parameters for Macro Read_StringArray. 2 Parameters required"
        Exitm
    #endif
    #if(Prm_1 != String) && (Prm_1 != Byte_Array_Ptr)       ' Ensure the first parameter is a string or a byte array
        #error "Parameter 1 must be a String Array for Macro Read_StringArray"
        Exitm
    #endif

    #if((Prm_2 == Num8) || (Prm_2 == Num16) || (Prm_2 == Num32))
        Lfsr 1,(pStringName + (pIndex * pStringName#v(99))) ' Point FSR1L/H to the address of the string array
    #endif
    #if((Prm_2 == Byte) || (Prm_2 == Word) || (Prm_2 == Dword) || (Prm_2 == SByte) || (Prm_2 == SWord) || (Prm_2 == SDword))
        Byte_Wreg pIndex
        Lfsr 1,(pStringName)                                ' Point FSR1L/H to the address of the string array
        Mullw (pStringName#v(99))                           ' \
        Movf PRODL,w                                        ' |
        Addwf FSR1L,f                                       ' | Calculate the position of the string array within the main STRING
        Movf PRODH,w                                        ' |
        Addwfc FSR1H,f                                      ' /
    #endif
'
' FSR1L and FSR1H are now pointing to the correct string within the array
'
    Endm

'--------------------------------------------------------------------------------------
' Read from a code memory String array
' The calculation for accessing the elements is: -
' TBLPTR = pStringName
' PROD = bIndex * string's length
' TBLPTR = TBLPTR + PROD

' Syntax:
' Dest_String = Read_StringArray, pStringLength, [pIndex]

' Where:-
' Dest_String is the destination string to hold the string extracted from the array
' pStringName is the name of the array in question
' pIndex is the index value of the string array
'
Read_CodeStringArray Macro- pCodeStringName, pIndex \Label

    #if(Prm_Count != 2)
        #error "Incorrect amount of parameters for Macro Read_StringArray. 2 Parameters required"
        Exitm
    #endif
    #if(Prm_1 != Label) && (Prm_1 != Byte_Array_Ptr)       ' Ensure the first parameter is a label name
        #error "Parameter 1 must be a Label name for Macro Read_CodeStringArray"
        Exitm
    #endif

    #if((Prm_2 == Num8) || (Prm_2 == Num16) || (Prm_2 == Num32))
        Movlw ((pCodeStringName + (pIndex * pCodeStringName#v(98))) & 255)
        Movwf TBLPTRL
        Movlw (((pCodeStringName + (pIndex * pCodeStringName#v(98))) >> 8) & 255)
        Movwf TBLPTRH
    #endif
    #if((Prm_2 == Byte) || (Prm_2 == Word) || (Prm_2 == Dword) || (Prm_2 == SByte) || (Prm_2 == SWord) || (Prm_2 == SDword))
        Movlw (pCodeStringName & 255)                       ' \
        Movwf TBLPTRL                                       ' |
        Movlw ((pCodeStringName >> 8) & 255)                ' | Point FSR1L/H to the address of the string array
        Movwf TBLPTRH                                       ' /
        Byte_Wreg pIndex
        Mullw (pCodeStringName#v(98))                       ' \
        Movf PRODL,w                                        ' |
        Addwf TBLPTRL,f                                     ' | Calculate the position of the string array within the main STRING
        Movf PRODH,w                                        ' |
        Addwfc TBLPTRH,f                                    ' /
    #endif
'
' TBLPTRL and TBLPTRH are now pointing to the correct string within the array
'
    #ifdef Read_CodeStringArray_Return
        #if Return_Type == String
            '
            ' FSR0L and FSR0H have already been loaded with the assignment string's address prior to the macro being used
            '
            GoSub _Read_CodeStringArray
        #else
            #error "Assignment must be a String variable for Macro Read_CodeStringArray"
        #endif
    #endif
    Endm
'
' Read from code memory into the assignment string
'
#ifMacro- Read_CodeStringArray
_Read_CodeStringArray:
    Clrf EECON1                                         ' \ Setup for reading from code memory
    Bsf EECON1,EEPGD                                    ' /
    Tblrd*+
    Movf TABLAT,w
    Bz $ + 6
    Movwf POSTINC0
    Bra $ - 8
    Return
#endIfMacro-
'--------------------------------------------------------------------------------------
_StringArray_Main_:
$endif  ' _STRINGARRAY_INC


